"""Command to clean old CSP reports."""
from datetime import timedelta

from django.core.management.base import BaseCommand, CommandError
from django.utils.encoding import force_str

from cspreports.models import get_report_model
from cspreports.utils import get_midnight, parse_date_input

CSPReport = get_report_model()

DEFAULT_OFFSET = 7


class Command(BaseCommand):
    help = "Delete old CSP reports."

    def add_arguments(self, parser):
        """Parse command arguments."""
        parser.add_argument(
            'limit', nargs='?',
            help="The date until which the reports be deleted. By defalt {} days ago.".format(DEFAULT_OFFSET))

    def handle(self, **options):
        verbosity = options['verbosity']

        limit = options['limit']
        if limit:
            try:
                limit = parse_date_input(limit)
            except ValueError as err:
                raise CommandError(force_str(err))
        else:
            limit = get_midnight() - timedelta(days=DEFAULT_OFFSET)

        CSPReport.objects.filter(created__lt=limit).delete()
        if verbosity >= 2:
            self.stdout.write("Deleted all reports created before {}.".format(limit))



from django.conf import settings


class Settings:
    """
    Shadow Django's settings with a little logic
    """

    @property
    def EMAIL_ADMINS(self):
        return getattr(settings, "CSP_REPORTS_EMAIL_ADMINS", True)

    @property
    def LOG(self):
        return getattr(settings, "CSP_REPORTS_LOG", True)

    @property
    def LOG_LEVEL(self):
        return getattr(settings, "CSP_REPORTS_LOG_LEVEL", "warning")

    @property
    def LOGGER_NAME(self):
        return getattr(settings, "CSP_REPORTS_LOGGER_NAME", "CSP Reports")

    @property
    def SAVE(self):
        return getattr(settings, "CSP_REPORTS_SAVE", True)

    @property
    def ADDITIONAL_HANDLERS(self):
        return getattr(settings, "CSP_REPORTS_ADDITIONAL_HANDLERS", [])

    @property
    def FILTER_FUNCTION(self):
        return getattr(settings, "CSP_REPORTS_FILTER_FUNCTION", None)

    @property
    def CSP_REPORT_MODEL(self):
        return getattr(settings, "CSP_REPORTS_MODEL", "cspreports.CSPReport")


app_settings = Settings()



from django.contrib import admin

from cspreports.models import get_report_model


CSPReport = get_report_model()


class CSPReportAdmin(admin.ModelAdmin):
    list_display = ('id', 'created', 'document_uri', 'blocked_uri')
    fields = ('created', 'modified', 'json_as_html')
    readonly_fields = ('created', 'modified', 'json_as_html')

    def json_as_html(self, instance):
        return "<br />" + instance.json_as_html()

    def document_uri(self, instance):
        return instance.data.get('csp-report', {}).get('document-uri')

    def blocked_uri(self, instance):
        return instance.data.get('csp-report', {}).get('blocked-uri')

    json_as_html.short_description = "Report"
    json_as_html.allow_tags = True


admin.site.register(CSPReport, CSPReportAdmin)



"""Print summary of CSP reports."""
from datetime import timedelta

from django.core.management.base import BaseCommand, CommandError
from django.utils.encoding import force_str

from cspreports.summary import DEFAULT_TOP, collect
from cspreports.utils import get_midnight, parse_date_input


def _parse_date_input(date_input, default_offset=0):
    """Parses a date input."""
    if date_input:
        try:
            return parse_date_input(date_input)
        except ValueError as err:
            raise CommandError(force_str(err))
    else:
        return get_midnight() - timedelta(days=default_offset)


class Command(BaseCommand):
    help = "Print summary of CSP reports."

    def add_arguments(self, parser):
        """Parse command arguments."""
        parser.add_argument(
            '--since',
            help="The start date of summary. By default yesterday.")
        parser.add_argument(
            '--to',
            help="The end date of summary. By default yesterday.")
        parser.add_argument(
            '--top', type=int, default=DEFAULT_TOP,
            help="Specifies the size of each section. By default %(default)s.")

    def handle(self, **options):
        since = _parse_date_input(options['since'], 1)
        to = _parse_date_input(options['to'], 1) + timedelta(days=1)
        top = options['top']

        summary = collect(since, to, top)
        self.stdout.write(summary.render())



""" Filters for use with the  setting. """
import json


def filter_browser_extensions(request):
    """ Filters out reports of CSP violations which are caused by browser extensions from common
        web browsers trying to inject resources.
    """
    json_str = request.body
    if isinstance(json_str, bytes):
        json_str = json_str.decode(request.encoding or "utf-8")
    try:
        report = json.loads(json_str).get("csp-report", {})
    except json.decoder.JSONDecodeError:
        return False
    # Ignore reports caused by browser extensions trying to load stuff
    src_file = report.get("source-file", "")
    ignored_prefixes = (
        "safari-extension://",
        "safari-web-extension://",
        "moz-extension://",
        "chrome-extension://",
    )
    if any(src_file.startswith(prefix) for prefix in ignored_prefixes):
        return False
    return True



"""Collect summary of CSP reports."""
from operator import attrgetter
from urllib.parse import urlsplit, urlunsplit

from django.template.loader import get_template

from cspreports.models import get_report_model


CSPReport = get_report_model()
DEFAULT_TOP = 10


def get_root_uri(uri):
    """Return root URI - strip query and fragment."""
    chunks = urlsplit(uri)
    return urlunsplit((chunks.scheme, chunks.netloc, chunks.path, '', ''))


class ViolationInfo:
    """Container for violation details.

    @ivar root_uri: A violation root URI.
    @ivar count: A number of violations related to the root URI.
    @ivar examples: List of violation examples.
    @ivar top: Maximal number of examples.
    """

    def __init__(self, root_uri, top=DEFAULT_TOP):
        self.root_uri = root_uri
        self.count = 0
        self.examples = []
        self.top = top

    def append(self, report):
        """Append a new CSP report."""
        assert report not in self.examples
        self.count += 1
        if len(self.examples) < self.top:
            self.examples.append(report)


class CspReportSummary:
    """CSP report summary.

    @ivar since: Date and time of the summary start
    @ivar to: Date and time of the summary end
    @ivar top: Size of each section

    @ivar total_count: Total number of CSP reports
    @ivar valid_count: Total number of valid reports
    @ivar invalid_count: Total number of invalid reports

    @ivar sources: List of top sources of violations ordered by descending count
    @type sources: List[ViolationInfo]
    @ivar blocks: List of top blocked URIs ordered by descending count
    @type blocks: List[ViolationInfo]
    @ivar invalid_reports: Examples of invalid CSP reports
    """

    def __init__(self, since, to, top=DEFAULT_TOP):
        self.since = since
        self.to = to
        self.top = top

        self.total_count = 0
        self.valid_count = 0
        self.sources = []
        self.blocks = []
        self.invalid_count = 0
        self.invalid_reports = ()

    def render(self):
        """Render the summary."""
        template = get_template('cspreports/summary.txt')
        return template.render(self.__dict__)


def collect(since, to, top=DEFAULT_TOP):
    """Collect the CSP report.

    @returntype: CspReportSummary
    """
    summary = CspReportSummary(since, to, top=top)
    queryset = CSPReport.objects.filter(created__range=(since, to))
    valid_queryset = queryset.filter(is_valid=True)
    invalid_queryset = queryset.filter(is_valid=False)

    summary.total_count = queryset.count()
    summary.valid_count = valid_queryset.count()

    # Collect sources
    sources = {}
    for report in valid_queryset:
        root_uri = get_root_uri(report.document_uri)
        info = sources.setdefault(root_uri, ViolationInfo(root_uri))
        info.append(report)
    summary.sources = sorted(sources.values(), key=attrgetter('count'), reverse=True)[:top]

    # Collect blocks
    blocks = {}
    for report in valid_queryset:
        root_uri = get_root_uri(report.blocked_uri)
        info = blocks.setdefault(root_uri, ViolationInfo(root_uri))
        info.append(report)
    summary.blocks = sorted(blocks.values(), key=attrgetter('count'), reverse=True)[:top]

    # Collect invalid reports
    summary.invalid_count = invalid_queryset.count()
    summary.invalid_reports = tuple(invalid_queryset[:top])

    return summary



from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from cspreports.utils import process_report


@require_POST
@csrf_exempt
def report_csp(request):
    """ The handler for browsers to send Content Security Policy violation reports to.
        The 'report-uri' in HTTP Content-Security-Policy headers should point to this view.
    """
    process_report(request)
    return HttpResponse('')



import json
import logging
from datetime import datetime
from importlib import import_module

from django.conf import settings
from django.core.mail import mail_admins
from django.utils.dateparse import parse_date
from django.utils.timezone import localtime, make_aware, now

from cspreports.models import get_report_model
from cspreports.conf import app_settings

CSPReport = get_report_model()

logger = logging.getLogger(app_settings.LOGGER_NAME)


def process_report(request):
    """ Given the HTTP request of a CSP violation report, log it in the required ways. """
    if not should_process_report(request):
        return
    if app_settings.EMAIL_ADMINS:
        email_admins(request)
    if app_settings.LOG:
        log_report(request)
    if app_settings.SAVE:
        save_report(request)
    if app_settings.ADDITIONAL_HANDLERS:
        run_additional_handlers(request)


def format_report(jsn):
    """ Given a JSON report, return a nicely formatted (i.e. with indentation) string.
        This should handle invalid JSON (as the JSON comes from the browser/user).
        We trust that Python's json library is secure, but if the JSON is invalid then we still
        want to be able to display it, rather than tripping up on a ValueError.
    """
    if isinstance(jsn, bytes):
        jsn = jsn.decode('utf-8')
    try:
        return json.dumps(json.loads(jsn), indent=4, sort_keys=True, separators=(',', ': '))
    except ValueError:
        return "Invalid JSON. Raw dump is below.\n\n" + jsn


def email_admins(request):
    user_agent = request.META.get('HTTP_USER_AGENT', '')
    report = format_report(request.body)
    message = "User agent:\n%s\n\nReport:\n%s" % (user_agent, report)
    mail_admins("CSP Violation Report", message)


def log_report(request):
    func = getattr(logger, app_settings.LOG_LEVEL)
    func("Content Security Policy violation: %s", format_report(request.body))


def save_report(request):
    message = request.body
    if isinstance(message, bytes):
        message = message.decode(request.encoding or settings.DEFAULT_CHARSET)

    report = CSPReport.from_message(message)
    report.user_agent = request.META.get('HTTP_USER_AGENT', '')
    report.save()


def run_additional_handlers(request):
    for handler in get_additional_handlers():
        handler(request)


_additional_handlers = None
_filter_function = None


def get_additional_handlers():
    """ Returns the actual functions from the dotted paths specified in ADDITIONAL_HANDLERS. """
    global _additional_handlers
    if not isinstance(_additional_handlers, list):
        handlers = []
        for name in app_settings.ADDITIONAL_HANDLERS:
            function = import_from_dotted_path(name)
            handlers.append(function)
        _additional_handlers = handlers
    return _additional_handlers


def parse_date_input(value):
    """Return datetime based on the user's input.

    @param value: User's input
    @type value: str
    @raise ValueError: If the input is not valid.
    @return: Datetime of the beginning of the user's date.
    """
    try:
        limit = parse_date(value)
    except ValueError:
        limit = None
    if limit is None:
        raise ValueError("'{}' is not a valid date.".format(value))
    limit = datetime(limit.year, limit.month, limit.day)
    if settings.USE_TZ:
        limit = make_aware(limit)
    return limit


def get_midnight():
    """Return last midnight in localtime as datetime.

    @return: Midnight datetime
    """
    limit = now()
    if settings.USE_TZ:
        limit = localtime(limit)
    return limit.replace(hour=0, minute=0, second=0, microsecond=0)


def import_from_dotted_path(name):
    module_name, function_name = name.rsplit('.', 1)
    return getattr(import_module(module_name), function_name)


def should_process_report(request):
    if not app_settings.FILTER_FUNCTION:
        return True
    global _filter_function
    if _filter_function is None:
        _filter_function = import_from_dotted_path(app_settings.FILTER_FUNCTION)
    return _filter_function(request)



# Generated by Django 1.10.7 on 2017-07-07 09:23
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cspreports', '0002_auto_20141011_1800'),
    ]

    operations = [
        migrations.AddField(
            model_name='cspreport',
            name='blocked_uri',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='column_number',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='disposition',
            field=models.CharField(blank=True, choices=[('enforce', 'enforce'), ('report', 'report')], max_length=10, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='document_uri',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='effective_directive',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='is_valid',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='line_number',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='original_policy',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='referrer',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='source_file',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='status_code',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cspreport',
            name='violated_directive',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cspreports', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cspreport',
            options={'ordering': ('-created',)},
        ),
    ]



# Generated by Django 1.10.8 on 2017-11-29 08:09
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cspreports', '0003_report_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='cspreport',
            name='user_agent',
            field=models.TextField(blank=True),
        ),
    ]



# STANDARD LIB
import json

from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.db import models
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.apps import apps

from cspreports.conf import app_settings

DISPOSITIONS = (
    ('enforce', 'enforce'),
    ('report', 'report'),
)

# Map of required CSP report fields to model fields
REQUIRED_FIELDS = (
    ('document-uri', 'document_uri'),
    ('referrer', 'referrer'),
    ('blocked-uri', 'blocked_uri'),
    ('violated-directive', 'violated_directive'),
    ('original-policy', 'original_policy'),
)
# Map of optional (CSP >= 2.0) CSP report fields to model fields
OPTIONAL_FIELDS = (
    ('disposition', 'disposition'),
    ('effective-directive', 'effective_directive'),
    ('source-file', 'source_file'),
    ('status-code', 'status_code'),
    ('line-number', 'line_number'),
    ('column-number', 'column_number'),
)


class CSPReportBase(models.Model):
    """Represents a CSP violation report.

    @ivar created: Date and time of report creation.
    @ivar modified: Date and time of last modification.
    @ivar json: Raw CSP report
    @ivar is_valid: Whether the CSP report is valid.

    All report fields are NULL is report is invalid.
    Report fields are NULL if the report follows the rules of lower protocol level.

    CSP 1.0 fields:
    @ivar document_uri: The URI of protected resource.
    @ivar referrer: The referrer attribute of the protected resource.
    @ivar blocked_uri: The URI of the resource that was prevented from loading due to the policy violation.
    @ivar violated_directive: The policy directive that was violated.
    @ivar original_policy: The original policy as received by the user-agent.

    CSP 2.0 fields
    @ivar effective_directive: The name of the policy directive that was violated.
    @ivar status_code: The status-code of the HTTP response that contained the protected resource.
    @ivar source_file: The URL of the resource where the violation occurred.
    @ivar line_number: The line number in source-file on which the violation occurred.
    @ivar column_number: The column number in source-file on which the violation occurred.

    CSP 3.0 fields
    @ivar disposition: The disposition of violation's policy.
    """

    class Meta:
        ordering = ('-created',)
        abstract = True

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    user_agent = models.TextField(blank=True)
    json = models.TextField()
    is_valid = models.BooleanField(default=False)
    # Individual report fields - use `TextField` because there are no limits by any specification for these fields.
    document_uri = models.TextField(blank=True, null=True)
    referrer = models.TextField(blank=True, null=True)
    blocked_uri = models.TextField(blank=True, null=True)
    violated_directive = models.TextField(blank=True, null=True)
    original_policy = models.TextField(blank=True, null=True)
    effective_directive = models.TextField(blank=True, null=True)
    status_code = models.PositiveSmallIntegerField(blank=True, null=True)
    source_file = models.TextField(blank=True, null=True)
    line_number = models.PositiveIntegerField(blank=True, null=True)
    column_number = models.PositiveIntegerField(blank=True, null=True)
    disposition = models.CharField(max_length=10, blank=True, null=True, choices=DISPOSITIONS)

    @property
    def nice_report(self):
        """Return a nicely formatted original report."""
        if not self.json:
            return '[no CSP report data]'
        try:
            data = json.loads(self.json)
        except ValueError:
            return "Invalid CSP report: '{}'".format(self.json)
        if 'csp-report' not in data:
            return 'Invalid CSP report: ' + json.dumps(data, indent=4, sort_keys=True, separators=(',', ': '))
        return json.dumps(data['csp-report'], indent=4, sort_keys=True, separators=(',', ': '))

    def __str__(self):
        return self.nice_report

    @classmethod
    def from_message(cls, message):
        """Creates an instance from CSP report message.

        If the message is not valid, the result will still have as much fields set as possible.

        @param message: JSON encoded CSP report.
        @type message: text
        """
        self = cls(json=message)
        try:
            decoded_data = json.loads(message)
        except ValueError:
            # Message is not a valid JSON. Return as invalid.
            return self
        try:
            report_data = decoded_data['csp-report']
        except KeyError:
            # Message is not a valid CSP report. Return as invalid.
            return self

        is_valid = True

        fields = REQUIRED_FIELDS + OPTIONAL_FIELDS
        for json_field_name, django_field_name in fields:
            value = report_data.get(json_field_name)
            # Try to pass the value through as much of Django's coercion/cleaning as possible, but
            # if the data is not entirely valid that's not a reason to not still save the report
            # if we can; even if the data isn't perfect, some information is better than none.
            to_python = cls._meta.get_field(django_field_name).to_python
            clean = cls._meta.get_field(django_field_name).clean
            try:
                value = to_python(value)
                # If the first conversion step worked, then store the value, even if the next
                # conversion fails
                setattr(self, django_field_name, value)
                value = clean(value, model_instance=self)
                setattr(self, django_field_name, value)
            except ValidationError:
                is_valid = False

        # Go through the REQUIRED_FIELDS list and make sure that a valid value was supplied for
        # each of them. Note that for str-type fields we treat a value of "" as valid, but if the
        # field is missing entirely (i.e. is None) we treat that as invalid. I (Adam) am not sure
        # if that's the correct way to do it, but it appears to be how we're doing it. Also note
        # that our model fields have null=True for these required fields, and if we took that off
        # then this block of code might be unnecessary, as the block of code above would catch the
        # ValidationError and set `is_valid` to False for us.
        for field, django_field_name in REQUIRED_FIELDS:
            field_value = getattr(self, django_field_name)
            if field_value is None:
                is_valid = False
                break
        self.is_valid = is_valid

        return self

    @property
    def data(self):
        """ Returns self.json loaded as a python object. """
        try:
            data = self._data
        except AttributeError:
            data = self._data = json.loads(self.json)
        return data

    def json_as_html(self):
        """ Print out self.json in a nice way. """

        # To avoid circular import
        from cspreports import utils

        formatted_json = utils.format_report(self.json)
        return mark_safe("<pre>\n%s</pre>" % escape(formatted_json))


class CSPReport(CSPReportBase):
    pass


def get_report_model():
    model_string = app_settings.CSP_REPORT_MODEL
    try:
        return apps.get_model(model_string, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "CSP_REPORTS_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "CSP_REPORTS_MODEL refers to model '%s' that has not been installed"
            % model_string
        )



#!/usr/bin/env python
from setuptools import find_packages, setup
import os

PACKAGES = find_packages()
REQUIREMENTS = ['django >=2.2,<5.0']
TEST_REQUIREMENTS = ['coverage']
EXTRAS_REQUIRE = {
    'quality': ['isort', 'flake8'],
    'test': TEST_REQUIREMENTS,
}
CLASSIFIERS = ['License :: OSI Approved :: MIT License',
               'Framework :: Django',
               'Programming Language :: Python',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3 :: Only',
               'Programming Language :: Python :: 3.4',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3.9',
               'Programming Language :: Python :: 3.10',
               'Programming Language :: Python :: 3.11']


DESCRIPTION = (
    "A Django app for handling reports from web browsers of violations of your website's "
    "HTTP Content Security Policy."
)
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), "README.md")).read()

setup(
    name='django-csp-reports',
    version="v1.9.1",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Adam Alton',
    author_email='adamalton@gmail.com',
    url='https://github.com/adamalton/django-csp-reports',
    packages=PACKAGES,
    include_package_data=True,
    python_requires='>=3.4',
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
    test_suite='runtests.runtests',
    keywords=['django', 'csp', 'content security policy'],
    classifiers=CLASSIFIERS,
)
